import aiinterface.AIInterface;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Vector;
import aiinterface.CommandCenter;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;
import enumerate.Action;
import enumerate.State;

public class MCTS {
	
	private FrameData mainFrameData;
	private FrameData simulatorAheadFrameData;
	
	private LinkedList<Action> my_Actions;
	private LinkedList<Action> opp_Actions;
	
	private CharacterData my_characterData;
	private CharacterData opp_characterData;
	
	private static final int FRAME_AHEAD = 14;
	
	private ArrayList<MotionData> myMotion;
	private ArrayList<MotionData> oppMotion;
	
	private Action[] actionAir;
	private Action[] actionGround;
	
	private Action[] actionsToPerform;
	
	private Action spSkill;
	private Node rootNode;
	
	public static final boolean DEBUG_MODE = false;
	
	void Initialize(GameData gd, boolean playerNumber, Action[] actionsAir, Action[] actionsGround, Action sp) {
		this.my_Actions = new LinkedList<Action>();
		this.opp_Actions = new LinkedList<Action>();
		
		this.actionAir = actionsAir;
		this.actionGround = actionsGround;
		this.spSkill = sp;
	    
	    myMotion = gd.getMotionData(playerNumber);
	    oppMotion = gd.getMotionData(!playerNumber);
	}
	
	Action subProcesing(Simulator sim, FrameData frameData, boolean playerNumber, GameData gameData, CommandCenter commandCenter) {
		Prepare(sim, frameData, playerNumber);
		rootNode = new Node(simulatorAheadFrameData, null, my_Actions, opp_Actions, gameData, playerNumber, commandCenter);
		rootNode.createNode();
		
		Action bestAction = rootNode.mcts();
		if(DEBUG_MODE) {
			rootNode.printNode(rootNode);
		}
		
		return bestAction;
	}
	
	void Prepare(Simulator sim, FrameData frameData, boolean playerNumber) {
		simulatorAheadFrameData = sim.simulate(frameData, playerNumber, null, null, FRAME_AHEAD);
		
		my_characterData = simulatorAheadFrameData.getCharacter(playerNumber);
		opp_characterData = simulatorAheadFrameData.getCharacter(!playerNumber);
		
		setMyAction();
		setOppAction();
		
	}
	
	public void setMyAction() {
    my_Actions.clear();

    int energy = my_characterData.getEnergy();

    if (my_characterData.getState() == State.AIR) {
      for (int i = 0; i < actionAir.length; i++) {
        if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          my_Actions.add(actionAir[i]);
        }
      }
    } else {
      if (Math.abs(myMotion.get(Action.valueOf(spSkill.name()).ordinal())
          .getAttackStartAddEnergy()) <= energy) {
        my_Actions.add(spSkill);
      }

      for (int i = 0; i < actionGround.length; i++) {
        if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          my_Actions.add(actionGround[i]);
        }
      }
    }
}

	  public void setOppAction() {
	  opp_Actions.clear();
	  int energy = opp_characterData.getEnergy();
	  
	  if (opp_characterData.getState() == State.AIR) {
		  for (int i = 0; i < actionAir.length; i++) {
			  if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
					  .getAttackStartAddEnergy()) <= energy) {
				  opp_Actions.add(actionAir[i]);
        }
      }
    } else {
      if (Math.abs(oppMotion.get(Action.valueOf(spSkill.name()).ordinal())
          .getAttackStartAddEnergy()) <= energy) {
        opp_Actions.add(spSkill);
      }

      for (int i = 0; i < actionGround.length; i++) {
        if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          opp_Actions.add(actionGround[i]);
        }
      }
    }
  }
}
